from os import read, close, fsencode, fsdecode
from struct import unpack
from fcntl import ioctl
from select import poll
from termios import FIONREAD
from time import sleep
from ctypes import cdll, c_int, POINTER
from errno import errorcode, EINTR


libc = cdll.LoadLibrary('libc.so.6')
libc.__errno_location.restype = POINTER(c_int)
_inotify = dict(
    init = libc.inotify_init,
    add = libc.inotify_add_watch,
    rm = libc.inotify_rm_watch,
)


def check(ret, ignore=None):
    if ret<0:
        errno = libc.__errno_location().contents.value
        if errno!=ignore:
            raise OSError(errno,  errorcode[errno])


def inotify(fct, *args):
    while True:
        val = _inotify[fct](*args)
        check(val, ignore=EINTR)
        return val


class Inotify:
    def __init__(self):
        self.ifd = inotify('init')
        check(self.ifd)
        poller = poll()
        poller.register(self.ifd)
        self.poll = poller.poll

    def read(self, timeout=None):
        '''
            timeout : ms ; or block
        '''
        events = self.poll(timeout)
        if not events:
            return

        size_int = c_int()
        ioctl(self.ifd, FIONREAD, size_int)
        size = size_int.value
        data = read(self.ifd, size)
        deb = 0
        while deb < size:
            fin = deb+16
            wd, mask, cookie, name_len = unpack('iIII', data[deb:fin])
            deb, fin = fin, fin+name_len
            name = unpack('%ds' % name_len, data[deb:fin])
            name = fsdecode(name[0].rstrip(b'\0'))
            deb = fin
            yield wd, mask, cookie, name

    def add_watch(self, path, mask):
        if not isinstance(path, bytes):
            path = fsencode(path)
        w = inotify('add', self.ifd, path, mask)
        check(w)
        return w

    def rm_watch(self, wd):
        r = inotify('rm', self.ifd, wd)
        check(r)

    def close(self):
        close(self.ifd)


FLAGS = dict(
    ACCESS      = 0x00000001, # IN_ACCESS
    MODIFY      = 0x00000002, # IN_MODIFY
    ATTRIB      = 0x00000004, # IN_ATTRIB
    WRITE       = 0x00000008, # IN_CLOSE_WRITE
    CLOSE       = 0x00000010, # IN_CLOSE_NOWRITE
    OPEN        = 0x00000020, # IN_OPEN
    MOVED_FROM  = 0x00000040, # IN_MOVED_FROM
    MOVED_TO    = 0x00000080, # IN_MOVED_TO
    CREATE      = 0x00000100, # IN_CREATE
    DELETE      = 0x00000200, # IN_DELETE
    DELETE_SELF = 0x00000400, # IN_DELETE_SELF
    MOVE_SELF   = 0x00000800, # IN_MOVE_SELF
    UNMOUNT     = 0x00002000, # IN_UNMOUNT
    Q_OVERFLOW  = 0x00004000, # IN_Q_OVERFLOW
    IGNORED     = 0x00008000, # IN_IGNORED
    ONLYDIR     = 0x01000000, # IN_ONLYDIR
    DONT_FOLLOW = 0x02000000, # IN_DONT_FOLLOW
    MASK_ADD    = 0x20000000, # IN_MASK_ADD
    ISDIR       = 0x40000000, # IN_ISDIR
    ONESHOT     = 0x80000000, # IN_ONESHOT
)


def mask_str(mask):
    return ' | '.join(name for name, val in FLAGS.items() if val & mask)




from os import listdir
from os.path import join, normpath, isdir
from threading import Thread, RLock
from queue import Queue
from pathlib import Path
from time import sleep

from .inotify import Inotify, FLAGS, mask_str


globals().update(FLAGS)
ALL = ACCESS|MODIFY|ATTRIB|WRITE|CLOSE|OPEN|CREATE|DELETE


class Event:
    def __init__(self, wd, watcher, mask, cookie, name):
        self.wd = wd
        self.watcher = watcher
        self.mask = mask
        self.cookie = cookie
        self.name = name

    @property
    def watch(self):
        return self.watcher.watches[self.wd]

    @property
    def path(self):
        path = self.watch.path
        path = join(path, self.name) if self.name else path
        return Path(path)

    def __str__(self):
        return '{} -> {}'.format(mask_str(self.mask), self.path)


class Watch:
    def __init__(self, wd, watcher, mask, path, callback, recur=False, _parent=None):
        self.wd = wd
        self.watcher = watcher
        self.mask = mask
        self.path = normpath(path)
        self.callback = callback
        self.recur = recur
        self._parent =_parent
        if isdir(self.path) and self.recur:
            for name in listdir(self.path):
                path =  join(self.path, name)
                if isdir(path):
                    self.watcher.add(
                        path, mask, callback, recur=True, _parent=self
                    )

    def __str__(self):
        return 'Watch {}.{} {} {} -> {}'.format(
            self.watcher.inotify.ifd,
            self.wd,
            self.path,
            mask_str(self.mask),
            self.callback.__name__,
        )


def default(event):
    if event.mask not in (ISDIR|OPEN, ISDIR|CLOSE): print(event)


class Watcher:
    def __init__(self):
        self.inotify = Inotify()
        self.watches = {}
        self.queue = Queue()
        self.lock = RLock()
        for t in (self._push, self._pull):
            t = Thread(target=t)
            t.setDaemon(True)
            t.start()

    def _push(self):
        while True:
            for wd, mask, cookie, name in self.inotify.read():
                self.queue.put(
                    Event(wd, self, mask, cookie, name)
                )

    def _pull(self):
        while True:
            event = self.queue.get()
            mask = event.mask
            with self.lock:
                watch = self.watches.get(event.wd)
            watch.callback(event)
            if mask&ISDIR and mask&CREATE and watch.recur:
                self.add(
                    event.path,
                    watch.mask,
                    watch.callback,
                    recur = True,
                    _parent = watch,
                )
            if mask&IGNORED:
                # text editors (vim, ...) -> IGNORE ???
                self.rem(watch)
                self.add(
                    path = watch.path,
                    mask = watch.mask,
                    callback = watch.callback,
                    recur = watch.recur,
                    _parent = watch._parent,
                )

    def add(self, path, mask=ALL, callback=default, recur=False, _parent=None):
        with self.lock:
            wd = self.inotify.add_watch(path, mask)
            watch = Watch(wd, self, mask, path, callback, recur, _parent)
            self.watches[wd] = watch
        return watch

    def rem(self, watch):
        with self.lock:
            for w in list(self.watches.values()):
                if w._parent==watch:
                    self.rem(w)
            wd = watch.wd
            self.watches.pop(wd)

    def close(self):
        self.inotify.close()

    def __enter__(self):
        return self

    def __exit__(self, typ, val, tbk):
        try:
            while True: sleep(11)
        except KeyboardInterrupt:
            print
        finally:
            self.close()




from os.path import exists
from string import Template
from subprocess import run
from sys import argv

from pyfsnotif import Watcher, MODIFY


def on_change_run(path, cmd):

    command = Template(cmd.replace('%','$path')).substitute(path=path).split(' ')

    def do_cmd(event):
        run(command)

    with Watcher() as w:
        w.add(path, MODIFY, do_cmd)


def on_change():
    path, *cmd = argv[1:]
    cmd = ' '.join(cmd)
    if exists(path):
        if '%' in cmd:
            on_change_run(path, cmd)
            return
    print(f'''Usage:
    on_change {{path}} {{cmd}}

    Use % in {{cmd}} as a placeholder for {{path}}.

    ex : on_change readme.txt cat %

    Use ^C to stop.
    ''')



from os.path import join, dirname, abspath

from setuptools import setup, find_packages

from pyfsnotif import __version__


curdir = abspath(dirname(__file__))
readme = open(join(curdir, 'README.rst')).read()

setup(
    name             = 'pyfsnotif',
    version          = __version__,
    description      = 'inotifier',
    long_description = readme,
    keywords         = ['utility', ],
    url              = 'https://gitlab.com/dugres/pyfsnotif/tree/stable',
    author           = 'Louis RIVIERE',
    author_email     = 'louis@riviere.xyz',
    license          = 'MIT',
    classifiers      = [
        'Development Status :: 4 - Beta',
        'Intended Audience :: Developers',
        'Topic :: Software Development :: Testing',
        'License :: OSI Approved :: MIT License',
        'Programming Language :: Python :: 3',
    ],
    package_dir = {
        'pyfsnotif': 'pyfsnotif',
    },
    packages = [
        'pyfsnotif',
    ],
    entry_points = dict(
        console_scripts = (
            'on_change=pyfsnotif.command:on_change',
        ),
    ),
)
